@@ -41,6 +41,7 @@
 #include "volumes.h"
 #include "transaction.h"
 #include "utils.h"
+#include "version.h"
 
 static u64 parse_size(char *s)
 {
@@ -68,8 +69,8 @@ static u64 parse_size(char *s)
 	return atol(s) * mult;
 }
 
-static int make_root_dir(int fd, const char *device_name) {
-	struct btrfs_root *root;
+static int make_root_dir(struct btrfs_root *root)
+{
 	struct btrfs_trans_handle *trans;
 	struct btrfs_key location;
 	u64 bytes_used;
@@ -77,12 +78,6 @@ static int make_root_dir(int fd, const char *device_name) {
 	u64 chunk_size = 0;
 	int ret;
 
-	root = open_ctree_fd(fd, device_name, 0, O_RDWR);
-
-	if (!root) {
-		fprintf(stderr, "ctree init failed\n");
-		return -1;
-	}
 	trans = btrfs_start_transaction(root, 1);
 	bytes_used = btrfs_super_bytes_used(&root->fs_info->super_copy);
 
@@ -118,7 +113,6 @@ static int make_root_dir(int fd, const char *device_name) {
 				     chunk_start, chunk_size);
 	BUG_ON(ret);
 
-	// ret = btrfs_make_block_group(trans, root, 0, 1);
 	ret = btrfs_make_root_dir(trans, root->fs_info->tree_root,
 			      BTRFS_ROOT_TREE_DIR_OBJECTID);
 	if (ret)
@@ -131,18 +125,17 @@ static int make_root_dir(int fd, const char *device_name) {
 	ret = btrfs_insert_dir_item(trans, root->fs_info->tree_root,
 			"default", 7,
 			btrfs_super_root_dir(&root->fs_info->super_copy),
-			&location, BTRFS_FT_DIR);
+			&location, BTRFS_FT_DIR, 0);
 	if (ret)
 		goto err;
 
 	ret = btrfs_insert_inode_ref(trans, root->fs_info->tree_root,
 			     "default", 7, location.objectid,
-			     BTRFS_ROOT_TREE_DIR_OBJECTID);
+			     BTRFS_ROOT_TREE_DIR_OBJECTID, 0);
 	if (ret)
 		goto err;
 
 	btrfs_commit_transaction(trans, root);
-	ret = close_ctree(root);
 err:
 	return ret;
 }
@@ -239,14 +232,46 @@ static int create_raid_groups(struct btrfs_trans_handle *trans,
 	return 0;
 }
 
+static int create_data_reloc_tree(struct btrfs_trans_handle *trans,
+				  struct btrfs_root *root)
+{
+	struct btrfs_key location;
+	struct btrfs_root_item root_item;
+	struct extent_buffer *tmp;
+	u64 objectid = BTRFS_DATA_RELOC_TREE_OBJECTID;
+	int ret;
+
+	ret = btrfs_copy_root(trans, root, root->node, &tmp, objectid);
+	BUG_ON(ret);
+
+	memcpy(&root_item, &root->root_item, sizeof(root_item));
+	btrfs_set_root_bytenr(&root_item, tmp->start);
+	btrfs_set_root_level(&root_item, btrfs_header_level(tmp));
+	btrfs_set_root_generation(&root_item, trans->transid);
+	free_extent_buffer(tmp);
+
+	location.objectid = objectid;
+	location.type = BTRFS_ROOT_ITEM_KEY;
+	location.offset = trans->transid;
+	ret = btrfs_insert_root(trans, root->fs_info->tree_root,
+				&location, &root_item);
+	BUG_ON(ret);
+	return 0;
+}
+
 static void print_usage(void)
 {
 	fprintf(stderr, "usage: mkfs.btrfs [options] dev [ dev ... ]\n");
 	fprintf(stderr, "options:\n");
+	fprintf(stderr, "\t -A --alloc-start the offset to start the FS\n");
 	fprintf(stderr, "\t -b --byte-count total number of bytes in the FS\n");
+	fprintf(stderr, "\t -d --data data profile, raid0, raid1, raid10 or single\n");
 	fprintf(stderr, "\t -l --leafsize size of btree leaves\n");
-	fprintf(stderr, "\t -n --nodesize size of btree leaves\n");
+	fprintf(stderr, "\t -L --label set a label\n");
+	fprintf(stderr, "\t -m --metadata metadata profile, values like data profile\n");
+	fprintf(stderr, "\t -n --nodesize size of btree nodes\n");
 	fprintf(stderr, "\t -s --sectorsize min block allocation\n");
+	fprintf(stderr, "%s\n", BTRFS_BUILD_VERSION);
 	exit(1);
 }
 
@@ -307,7 +332,7 @@ int main(int ac, char **av)
 	char *first_file;
 	u64 block_count = 0;
 	u64 dev_block_count = 0;
-	u64 blocks[6];
+	u64 blocks[7];
 	u64 alloc_start = 0;
 	u64 metadata_profile = BTRFS_BLOCK_GROUP_RAID1 | BTRFS_BLOCK_GROUP_DUP;
 	u64 data_profile = BTRFS_BLOCK_GROUP_RAID0;
@@ -352,6 +377,13 @@ int main(int ac, char **av)
 				break;
 			case 'b':
 				block_count = parse_size(optarg);
+				if (block_count < 256*1024*1024) {
+					fprintf(stderr, "File system size "
+						"%llu bytes is too small, "
+						"256M is required at least\n",
+						block_count);
+					exit(1);
+				}
 				zero_end = 0;
 				break;
 			default:
@@ -371,6 +403,9 @@ int main(int ac, char **av)
 	if (ac == 0)
 		print_usage();
 
+	printf("\nWARNING! - %s IS EXPERIMENTAL\n", BTRFS_BUILD_VERSION);
+	printf("WARNING! - see http://btrfs.wiki.kernel.org before using\n\n");
+
 	file = av[optind++];
 	ret = check_mounted(file);
 	if (ret < 0) {
@@ -393,8 +428,11 @@ int main(int ac, char **av)
 	if (block_count == 0)
 		block_count = dev_block_count;
 
-	for (i = 0; i < 6; i++)
-		blocks[i] = BTRFS_SUPER_INFO_OFFSET + leafsize * i;
+	blocks[0] = BTRFS_SUPER_INFO_OFFSET;
+	for (i = 1; i < 7; i++) {
+		blocks[i] = BTRFS_SUPER_INFO_OFFSET + 1024 * 1024 +
+			leafsize * i;
+	}
 
 	ret = make_btrfs(fd, file, label, blocks, block_count,
 			 nodesize, leafsize,
@@ -403,13 +441,15 @@ int main(int ac, char **av)
 		fprintf(stderr, "error during mkfs %d\n", ret);
 		exit(1);
 	}
-	ret = make_root_dir(fd, file);
+	root = open_ctree(file, 0, O_RDWR);
+	root->fs_info->alloc_start = alloc_start;
+
+	ret = make_root_dir(root);
 	if (ret) {
 		fprintf(stderr, "failed to setup the root directory\n");
 		exit(1);
 	}
-	root = open_ctree(file, 0, O_RDWR);
-	root->fs_info->alloc_start = alloc_start;
+
 	trans = btrfs_start_transaction(root, 1);
 
 	if (ac == 0)
@@ -461,12 +501,17 @@ int main(int ac, char **av)
 raid_groups:
 	ret = create_raid_groups(trans, root, data_profile,
 				 metadata_profile);
+	BUG_ON(ret);
+
+	ret = create_data_reloc_tree(trans, root);
+	BUG_ON(ret);
 
 	printf("fs created label %s on %s\n\tnodesize %u leafsize %u "
 	    "sectorsize %u size %s\n",
 	    label, first_file, nodesize, leafsize, sectorsize,
 	    pretty_sizes(btrfs_super_total_bytes(&root->fs_info->super_copy)));
 
+	printf("%s\n", BTRFS_BUILD_VERSION);
 	btrfs_commit_transaction(trans, root);
 	ret = close_ctree(root);
 	BUG_ON(ret);
